<?php

declare(strict_types=1);

namespace Dedi\SyliusMessengerMonitorPlugin\Storage;

use Doctrine\Bundle\DoctrineBundle\Registry;
use Doctrine\ORM\EntityRepository;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Messenger\Envelope;
use Zenstruck\Collection;
use Zenstruck\Collection\Doctrine\ORM\EntityResult;
use Zenstruck\Messenger\Monitor\History\Model\ProcessedMessage;
use Zenstruck\Messenger\Monitor\History\Model\Results;
use Zenstruck\Messenger\Monitor\History\Specification;
use Zenstruck\Messenger\Monitor\History\Storage;

class ORMHistoryStorage implements Storage
{
    private EntityRepository $repository;

    public function __construct(
        private Storage $baseStorage,
        private Registry $registry,
        private string $entityClass,
    ) {
    }

    public function find(mixed $id): ?ProcessedMessage
    {
        return $this->baseStorage->find($id);
    }

    public function filter(Specification $specification): Collection
    {
        return $this->baseStorage->filter($specification);
    }

    public function purge(Specification $specification): int
    {
        return $this->baseStorage->purge($specification);
    }

    public function save(Envelope $envelope, Results $results, ?\Throwable $exception = null): void
    {
        $this->baseStorage->save($envelope, $results, $exception);
    }

    public function delete(mixed $id): void
    {
        $this->baseStorage->delete($id);
    }

    public function averageWaitTime(Specification $specification): ?float
    {
        return $this->baseStorage->averageWaitTime($specification);
    }

    public function averageHandlingTime(Specification $specification): ?float
    {
        return $this->baseStorage->averageHandlingTime($specification);
    }

    public function count(Specification $specification): int
    {
        return $this->baseStorage->count($specification);
    }

    public function availableMessageTypes(Specification $specification): Collection
    {
        $qb = $this->queryBuilderFor($specification->ignoreMessageType(), false)
            ->select('DISTINCT m.type')
            ->addSelect('m.id')
            ->orderBy('m.type', 'ASC')
        ;

        return (new EntityResult($qb))->asString();
    }

    private function queryBuilderFor(Specification $specification, bool $order = true): QueryBuilder
    {
        [$from, $to, $status, $messageType, $transport, $tags, $notTags, $sort, $runId] = \array_values($specification->toArray());

        $qb = $this->repository()->createQueryBuilder('m');

        if ($from !== null) {
            $qb->andWhere('m.finishedAt >= :from')->setParameter('from', $from);
        }

        if ($to !== null) {
            $qb->andWhere('m.finishedAt <= :to')->setParameter('to', $to);
        }

        if ($messageType !== null) {
            $qb->andWhere('m.type = :class')->setParameter('class', $messageType);
        }

        if ($transport !== null) {
            $qb->andWhere('m.transport = :transport')->setParameter('transport', $transport);
        }

        if ($runId !== null) {
            $qb->andWhere('m.runId = :run_id')->setParameter('run_id', $runId);
        }

        match ($status) {
            Specification::SUCCESS => $qb->andWhere('m.failureType IS NULL'),
            Specification::FAILED => $qb->andWhere('m.failureType IS NOT NULL'),
            null => null,
        };

        foreach ($tags as $i => $tag) {
            $qb->andWhere('m.tags LIKE :tag' . $i)->setParameter('tag' . $i, '%' . $tag . '%');
        }

        foreach ($notTags as $i => $notTag) {
            $expr = [] === $tags
                ? $qb->expr()->orX(
                    $qb->expr()->isNull('m.tags'),
                    $qb->expr()->notLike('m.tags', ':not_tag' . $i),
                )
                : 'm.tags NOT LIKE :not_tag' . $i;

            $qb->andWhere($expr)->setParameter('not_tag' . $i, '%' . $notTag . '%');
        }

        if ($order === true) {
            $qb->orderBy('m.finishedAt', \mb_strtoupper($sort));
        }

        return $qb;
    }

    private function repository(): EntityRepository
    {
        if (isset($this->repository)) {
            return $this->repository;
        }

        /** @var EntityRepository|null $repository */
        $repository = $this->registry->getRepository($this->entityClass);

        if (!$repository instanceof EntityRepository) {
            throw new \LogicException('Only the ORM is currently supported.');
        }

        return $this->repository = $repository;
    }
}
